/**************************************************************************
 * KMerlin logging feature
 * implemented by Bernard Blackham <bernard_AT_blackham.com.au>
 *
 * The plan is to automatically log conversations, as they occur.
 * Due to the nature of multiple-user conversations, the layout shall be
 * as follows:
 *
 *     * Each conversation window will be logged into a separate file.
 *       This file will be in a subdirectory of the loggingDir as:
 *         $LOGGINGDIR/YYYY/MM/$EPOCHTIME.$UNIQUE
 *       where $EPOCHTIME is the number of seconds since the epoch and
 *       $UNIQUE is a number to make sure the filename is unique.
 *     
 *     * Also in $LOGGINGDIR will be directories of the names of the
 *       people in conversations. eg, tatoonie10@hotmail.com will 
 *       become $LOGGINGDIR/tatoonie10_AT_hotmail.com
 *
 *       and under these directories will be symbolic links to the files
 *       created previously, so that every user involved in a chat will
 *       have a symbolic link to that chat file.
 *
 * Every chat window has a Logger class iff logging is enabled.
 * 
 * Makes sense? I hope so.
 *
 * The main functions here are:
 *   logEntry() - called when a line is added to a convo log
 *   logAddUser() - creates a symlink for the appropriate user
 *
 ***************************************************************************/

#include <unistd.h>

#include <qfile.h>
#include <qtextstream.h>

#include "chatmember.h"
#include "logger.h"

Logger::Logger(QString theLoggingDir) {
	// take the time and date and make a unique conversation
	// create the appropriate file handles et al.
	logging = false;
	loggingDir = theLoggingDir;
	if (theLoggingDir == QString::null) return;
	if (!createLogFileName()) return;
	logFile.setName(logFileName);
	if (!logFile.open(IO_WriteOnly | IO_Append | IO_Translate)) {
		qWarning("Couldn't open chat log file \"%s\".", 
				(const char *)logFileName);
		logErrorNoted = true;
	} else {
		logStream.setDevice(&logFile);
		logging = true;
		logErrorNoted = false; // we only whinge about logging errors the first
							   // time
	}
}

Logger::~Logger() {
	logFile.close();
}

bool Logger::createLogFileName() {
	// returns the absolute file name of a new logfile that is safe to
	// append to. It also creates the file empty.
	QDateTime dt = QDateTime::currentDateTime();
	
	QDir d(loggingDir);
	QString year; year.setNum(dt.date().year());
	QString month; month.setNum(dt.date().month());
	if (!safeCd(&d, year)) return false;
	if (!safeCd(&d, month)) return false;

	// filename is the time in seconds since epoch + unique identifier
	QDateTime epoch; epoch.setTime_t(0);
	
	int epochTime = epoch.secsTo(dt);
	int uniqueNum = 0;
	bool success = false;
	QString uniqueName;
	while (!success && uniqueNum != -1) {
		uniqueName.sprintf("%d.%d", epochTime, uniqueNum);
		QFileInfo fi(d, uniqueName);
		if (!fi.exists()) {
			// this is good.
			QFile f(fi.absFilePath());
			if (f.open(IO_WriteOnly)) {
				f.close();
				logFileName = fi.absFilePath();
				success = true;
			}
		}
		
		uniqueNum++;
		if (uniqueNum > 999) { // something seriously wrong!
			uniqueNum = -1;
			qWarning("Something seriously wrong. Can't create unique log file"
					 " in \"%s\"", (const char*)d.absPath());
			logErrorNoted = true;
			uniqueName = QString::null;
			return false;
		}
	}
	relLogFileName = year + "/" + month + "/" + uniqueName;
	return true;
}

QString Logger::memberDirName(QString handle) {
	// returns the name of a directory to put a certain user into
	QString dirName("");
	dirName.append(handle); // make sure it's not the original. else we're in deep trouble
	dirName.replace(QRegExp("@"), "_AT_");
	dirName.replace(QRegExp("^\\."), "_dot_");
	dirName.replace(QRegExp("~"), "_tilde_");
	return dirName;
}

bool Logger::safeCd(QDir* d, QString dn) {
	if (!d->cd(dn)) {
		QFileInfo fi(*d, dn);
		if (fi.exists()) {
			if (fi.isDir()) 
				qWarning("Cannot cd into \"%s\".", (const char*)d->absFilePath(dn));
			else
				qWarning("Cannot create directory \"%s\"\n"
						 "A file named \"%s\" already exists in \"%s\".",
						(const char *)d->absFilePath(dn),
						(const char *)dn,
						(const char *)d->path());
			return false;
		} else {
			if (!d->mkdir(dn)) {
				qWarning("Could not create directory \"%s\".",
						(const char *)d->absFilePath(dn));
				return false;
			} else {
				if (!d->cd(dn)) {
					qWarning("Something very sus creating directory \"%s\".",
							(const char *)d->absFilePath(dn));
				}
			}
		}
	}
	return true;
}

QString Logger::newEntryText() {
	// returns the string at the start of a new entry in a log
	// file. "[YYYYMMDD HHMMSS] "
	QDateTime dt = QDateTime::currentDateTime();
	QString dts;
	dts.sprintf("[%4d%02d%02d %02d%02d%02d] ",
			dt.date().year(),
			dt.date().month(),
			dt.date().day(),
			dt.time().hour(),
			dt.time().minute(),
			dt.time().second());
	return dts;
}

bool Logger::logEntry(QString text) {
	if (!logging) return false;
	logStream << newEntryText() << text << endl;
	logFile.flush();
	return true;
}

bool Logger::logAddUser(QString handle) {
	if (!logging) return false;
	// creates a symlink to a chat (note not terribly portable, but hey)
	QDir d(loggingDir);
	QString dn = memberDirName(handle);
	if (!safeCd(&d, dn)) return false;
	QFileInfo l((const char*)logFile.name());
	QFileInfo linkFile(d, l.fileName());

	QString linkPath; linkPath.sprintf("../%s", (const char *)relLogFileName);

	return symlink((const char*)linkPath, linkFile.absFilePath());
}
